/************************************************************************

  This file is a part of the wrapper program for launching scripts
  and programs in TeX Live on Windows. See readme.txt for more details.
  
  This file adds UAC support for programs that need it,
  in particular tlshell.

  It is written in 2019 by Siep Kroonenberg and is also
  placed in the Public Domain.
 
************************************************************************/

#include <windows.h>
#include <winnt.h>
#include <shellapi.h>
#include <shlobj.h>
// #include <wtypes.h>
#include <tchar.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

static TCHAR * instroot;
static char release[5];

#define FILE_EXISTS(f) ( GetFileAttributes(f) != INVALID_FILE_ATTRIBUTES )

// free and set to NULL
#define FREE0(p) { if( p ) { free( p ); p = NULL; }}

__declspec(dllimport) int dllwrunscript(
  HINSTANCE hInstance,
  HINSTANCE hPrevInstance,
  char *argline,
  int winshow 
 );

void die( TCHAR * s, int i ) {
  MessageBox( NULL, s, _T("ERROR!"),
      MB_ICONERROR | MB_TASKMODAL | MB_SETFOREGROUND );
  // PostQuitMessage( i ); // pointless: no message loop
  ExitProcess( i );
}

void * calloc_fatal( size_t nitems, size_t item_size ) {
  void * p;
  p = calloc( nitems, item_size );
  if ( !p ) ExitProcess( 1 );
  return p;
}

// [v]asprintf with automatic sizing and allocation
// adapted from source code of asprintf and vasprintf
int tl_vtasprintf( TCHAR ** ret, const TCHAR * fmt, va_list ap ) {
  FILE * fp;
  TCHAR * buf;
  int len;

  fp = fopen( "NUL", "w" );
  if ( ! fp ) {
    *ret = NULL;
    return -1;
  }
  len = _vftprintf( fp, fmt, ap );
  fclose( fp );
  if( len < 0 ) return -1;
  buf = calloc_fatal( len + 1, sizeof( TCHAR ));
  if ( _vstprintf( buf, fmt, ap ) != len ) {
    FREE0( buf );
    *ret = NULL;
    return -1;
  }
  *ret = buf;
  return len;
}

void tl_tasprintf( TCHAR ** ret, const TCHAR * fmt, ... ) {
  int len;
  va_list ap;
  va_start( ap, fmt );
  len = tl_vtasprintf( ret, fmt, ap );
  va_end( ap );
  if ( len < 0 ) die( _T("tl_tasprintf: Failure to construct string"), 1 );
}

void get_tl_release() {
  // read release from TL database
  TCHAR * dbfilename;
  char * cptr;
  int i, j, l;
  FILE * dbfile;
  char dbline[200];

  // calculate filename of TL database
  tl_tasprintf( &dbfilename, _T("%s\\tlpkg\\texlive.tlpdb"), instroot );
  // read start of database
  dbfile = _tfopen( dbfilename, _T("r") );
  if ( !dbfile ) {
    release[0] = '\0';
    return;
  }
  free( dbfilename );
  for ( i=0; i<200; i++ ) {
    memset( dbline, 0, sizeof( dbline ));
    cptr = fgets( dbline, sizeof( dbline ), dbfile );
    if( !cptr ) {
      // i.e. EOF or error
      fclose( dbfile );
      release[0] = '\0';
      return;
    }
    l = strlen( "depend release/" );
    if( !strncmp( "depend release/", dbline, l)) {
      // next 4 chars should be digits
      for( j=0; j<4; j++ ) {
        release[j] = dbline[l+j];
        if( release[j] < '0' || release[j] > '9' ) {
          memset( release, 0, sizeof( release ));
          break;
        }
      }
      release[4] = '\0';
      fclose( dbfile );
      return;
    }
  }
  // if we got here, release should still consist
  // of zeroes from the initial default initialization
  fclose( dbfile );
  return;
}

BOOL in_HKLM() {
  /*
  Check for uninstall reg entry in HKLM for current TL.

  On w64, registry entries under HKLM\software\wow6432node will be
  transparently aliased to corresponding entries under HKLM\software
  so for 32-bits programs no worries on this score.

  For just checking the existence of the uninstall registry key,
  we can get away with ascii/ansi/utf8.

  */
  long result;
  HKEY uninst_key;
  char keyname[] =
     "Software\\Microsoft\\windows\\currentversion\\uninstall\\TeXLivexxxx";

  if (!release[0]) return FALSE;
  strncpy( keyname + strlen( keyname ) - 4, release, 4 );

  result = RegOpenKeyExA(
    HKEY_LOCAL_MACHINE,
    keyname,
    0,
    KEY_QUERY_VALUE,
    &uninst_key);
  // An additional check: compare the path of the uninstaller
  // with the path of this program, using
  // the GetFileInformationByHandle API call, this time using TCHAR
  return ( result == ERROR_SUCCESS );
}

BOOL is_elevated() {

  BOOL ret = FALSE;

  // http://stackoverflow.com/questions/8046097/
  HANDLE ProcessHandle, TokenHandle = NULL;
  TOKEN_ELEVATION Elevation;
  DWORD cbsize = sizeof( TOKEN_ELEVATION );
  ProcessHandle = GetCurrentProcess(); // no need to close afterwards
  if( OpenProcessToken( ProcessHandle, TOKEN_QUERY, &TokenHandle )) {
    if( GetTokenInformation(
        TokenHandle,
        TokenElevation, // NOT a variable!
        &Elevation,
        sizeof( Elevation ),
        &cbsize )) {
      ret = Elevation.TokenIsElevated;
      if( TokenHandle ) CloseHandle( TokenHandle );
    }
  }
  return ret;
}

int APIENTRY WinMain( 
  HINSTANCE hInstance,
  HINSTANCE hPrevInstance,
  TCHAR *argline,
  int winshow ){

  // calculate installation root: grandparent of own directory
  // msdn is conflicted as to whether _wpgmptr is absolute;
  // don't use _get_wpgmptr
  TCHAR own_path[MAX_PATH], * tptr;
  int i, l;
  SHELLEXECUTEINFO shExecInfo;

  l = (int) GetModuleFileName( NULL, own_path, MAX_PATH );
  // length l excludes \0
  if ( l == 0 || GetLastError() != ERROR_SUCCESS )
    die( _T("wrunscript_UAC: cannot get own path"), 1 );
  // just copying, but let tl_tasprintf handle the allocation nuisance:
  tl_tasprintf( &instroot, _T("%s"), own_path );
  for (i=0; i<3; i++) {
    tptr = _tcsrchr( instroot, _T('\\' ));
    if ( !tptr ) die( _T("Did not find TeX Live installation"), 0 );
    *tptr = _T('\0');
  }

  get_tl_release(); // read TL release from release file
  if( ! release[0] ) {
    MessageBox( NULL,
       _T("Problem with installation: cannot determine release!"),
       _T("Error"), MB_OK | MB_ICONSTOP );
    return 1;
  }
  if ( ! in_HKLM() )
    return dllwrunscript( hInstance, hPrevInstance, argline, winshow );
  else if ( is_elevated() )
    return dllwrunscript( hInstance, hPrevInstance, argline, winshow );
  else {
    // this is the recommended way to get an elevation prompt(!)
    shExecInfo.cbSize = sizeof(SHELLEXECUTEINFO);
    shExecInfo.fMask = SEE_MASK_NO_CONSOLE;
    shExecInfo.hwnd = NULL;
    shExecInfo.lpVerb = _T("runas");
    shExecInfo.lpFile = own_path;
    shExecInfo.lpParameters = argline;
    shExecInfo.lpDirectory = NULL;
    shExecInfo.nShow = SW_HIDE;
    shExecInfo.hInstApp = NULL;
    if ( !ShellExecuteEx( &shExecInfo ))
      die( _T("Failed to restart elevated"), 1 );
    return 0;
  }
}
